""" BaseEmbeddings """

import copy

import pytorch_lightning as pl
import pytorch_lightning.core.lightning as lightning
import torch.nn as nn


class BaseEmbedding(lightning.LightningModule):
    """ All trainable embeddings must inherit from BaseEmbedding.

    """

    def __init__(self, model, criterion, optimizer, dataloader):
        """ Constructor

        Args:
            model: (torch.nn.Module)
            criterion: (torch.nn.Module)
            optimizer: (torch.optim.Optimizer)
            dataloader: (torch.utils.data.DataLoader)

        """

        super(BaseEmbedding, self).__init__()
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.dataloader = dataloader
        self.checkpoint = None

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y, _ = batch
        y_hat = self(x)
        loss = self.criterion(y_hat, y)
        tensorboard_logs = {'train_loss': loss}
        return {'loss': loss, 'log': tensorboard_logs}

    def configure_optimizers(self):
        return self.optimizer

    def train_dataloader(self):
        return self.dataloader

    def train_embedding(self, **kwargs):
        """ Train the model on the provided dataset.

        Args:
            **kwargs: pylightning_trainer arguments, examples include:
                min_epochs: (int) Minimum number of epochs to train
                max_epochs: (int) Maximum number of epochs to train
                gpus: (int) number of gpus to use

        Returns:
            A trained encoder, ready for embedding datasets.

        """
        trainer = pl.Trainer(**kwargs)
        trainer.fit(self)

        checkpoint_cb = trainer.checkpoint_callback
        try:
            self.checkpoint = checkpoint_cb.kth_best_model
        except:
            print('Warning: "kth_best_model" was deprecated. \
                    Using "kth_best_model_path" instead.')
            self.checkpoint = checkpoint_cb.kth_best_model_path

        return self

    def recycle(self, new_output_dim, layer=None):
        """Build a copy of the embedding model with a new output layer

        Args:
            new_output_dim: (int)
            layer: (int)

        Returns:
            Copy of the embedding model with new output layer
        """

        layer = -1 if layer is None else layer

        modules = []
        for module in self.model.features:
            module_copy = copy.deepcopy(module)
            modules.append(module_copy)

        output_dim = None
        if isinstance(modules[-1], nn.AdaptiveAvgPool2d):
            output_dim = modules[-2].out_channels
        elif isinstance(modules[-1], nn.Linear):
            output_dim = modules[-1].out_features
        else:
            msg = 'Could not determine output_size. Last layer is {}'
            msg = msg.format(type(modules[-1]))
            raise NotImplementedError(msg)

        modules.append(nn.Flatten())
        modules.append(nn.Linear(output_dim, new_output_dim))
        return nn.Sequential(*modules)

    def embed(self, *args, **kwargs):
        """Must be implemented by classes which inherit from BaseEmbedding.

        """
        raise NotImplementedError()



import random
import unittest

import torchvision
import torchvision.transforms as transforms

from boris.data import BaseCollateFunction
from boris.data import ImageCollateFunction


class TestDataCollate(unittest.TestCase):

    def create_batch(self, batch_size=16):
        rnd_images = torchvision.datasets.FakeData(size=batch_size)

        fnames = [f'img_{i}.jpg' for i in range(batch_size)]
        labels = [random.randint(0, 5) for i in range(batch_size)]

        batch = []

        for i in range(batch_size):
            batch.append((rnd_images[i][0], labels[i], fnames[i]))

        return batch

    def test_base_collate(self):
        batch = self.create_batch()
        transform = transforms.ToTensor()
        collate = BaseCollateFunction(transform)
        samples, labels, fnames = collate(batch)

        self.assertIsNotNone(collate)
        self.assertEqual(len(samples), len(labels), len(fnames))

    def test_image_collate(self):
        batch = self.create_batch()
        img_collate = ImageCollateFunction()
        samples, labels, fnames = img_collate(batch)

        self.assertIsNotNone(img_collate)
        self.assertEqual(len(samples), len(labels), len(fnames))



import os
import time
import random
import requests

SERVER_LOCATION = os.getenvb(b'BORIS_SERVER_LOCATION',
                             b'https://api.whattolabel.com').decode()


def _post_request(dst_url, data=None, json=None,
                  max_backoff=32, max_retries=5):

    counter = 0
    backoff = 1. + random.random() * 0.1
    success = False
    while not success:

        response = requests.post(dst_url, data=data, json=json)
        success = (response.status_code == 200)

        # exponential backoff
        if response.status_code in [500, 502]:
            time.sleep(backoff)
            backoff = 2*backoff if backoff < max_backoff else backoff
        # something went wrong
        elif not success:
            msg = f'Failed POST request to {dst_url} with status_code '
            msg += f'{response.status_code}.'
            raise RuntimeError(msg)

        counter += 1
        if counter >= max_retries:
            break

    if not success:
        msg = f'The connection to the server at {dst_url} timed out. '
        raise RuntimeError(msg)

    return response


def _put_request(dst_url, data,
                 max_backoff=32, max_retries=5):

    counter = 0
    backoff = 1. + random.random() * 0.1
    success = False
    while not success:

        response = requests.put(dst_url, data)
        success = (response.status_code == 200)

        # exponential backoff
        if response.status_code in [500, 502]:
            time.sleep(backoff)
            backoff = 2*backoff if backoff < max_backoff else backoff
        # something went wrong
        elif not success:
            msg = f'Failed PUT request to {dst_url} with status_code '
            msg += f'{response.status_code}.'
            raise RuntimeError(msg)

        counter += 1
        if counter >= max_retries:
            break

    if not success:
        msg = f'The connection to the server at {dst_url} timed out. '
        raise RuntimeError(msg)

    return response


def _get_request(dst_url, params=None,
                 max_backoff=32, max_retries=5):

    counter = 0
    backoff = 1. + random.random() * 0.1
    success = False
    while not success:

        response = requests.get(dst_url, params=params)
        success = (response.status_code == 200)

        # exponential backoff
        if response.status_code in [500, 502]:
            time.sleep(backoff)
            backoff = 2*backoff if backoff < max_backoff else backoff
        # something went wrong
        elif not success:
            msg = f'Failed GET request to {dst_url} with status_code '
            msg += f'{response.status_code}.'
            raise RuntimeError(msg)

        counter += 1
        if counter >= max_retries:
            break

    if not success:
        msg = f'The connection to the server at {dst_url} timed out. '
        raise RuntimeError(msg)

    return response


def create_initial_tag(dataset_id: str, token: str):
    """Makes empty post request to dataset to create initial tag

    Args:
        dataset_id: Identifier of the dataset
        token: The token for authenticating the request

    Raises:
        RuntimeError if creation of initial tag failed
    """
    payload = {
        'token': token
    }
    dst_url = f'{SERVER_LOCATION}/users/datasets/{dataset_id}/tags'
    response = _post_request(dst_url, json=payload)
    return response


def get_presigned_upload_url(filename: str,
                             dataset_id: str,
                             token: str) -> str:
    """Creates and returns a signed url to upload an image for a specific dataset

    Args:
        filename: Name of file used for database and cloud storage
        dataset_id: Identifier of the dataset
        token: The token for authenticating the request

    Returns:
        A string containing the signed url

    Raises:
        RuntimeError if requesting signed url failed
    """
    payload = {
        'sample': {
            'fileName': filename,
        },
        'token': token
    }
    dst_url = f'{SERVER_LOCATION}/users/datasets/{dataset_id}/samples'
    response = _post_request(dst_url, json=payload)
    data = response.json()
    signed_url = data['signedWriteUrl']
    return signed_url


def upload_file_with_signed_url(filename: str, url: str) -> bool:
    """Upload a file to the cloud storage using a signed URL

    Args:
        filename: Path to a file for upload
        url: Signed url for push

    Returns:
        A boolean value indicating successful upload
    """
    file = open(filename, 'rb')
    response = _put_request(url, data=file)
    file.close()
    return response


def upload_embedding(data: dict) -> bool:
    """Uploads embedding

    Args:
        data: Object with embedding data

    Returns:
        A boolean value indicating successful upload

    Raises:
        RuntimeError if upload was not successful
    """
    payload = {
        'embeddingName': data['embeddingName'],
        'embeddings': data['embeddings'],
        'token': data['token'],
        'append': data['append'],
    }
    dataset_id = data['datasetId']
    dst_url = f'{SERVER_LOCATION}/users/datasets/{dataset_id}/embeddings'
    response = _post_request(dst_url, json=payload)
    return response


def get_samples(dataset_id: str,
                token: str,
                tag_name: str = 'initial-tag'):
    # payload is the same for both requests
    payload = {
        'token': token
    }

    # get tag_id
    dst_url = f'{SERVER_LOCATION}/users/datasets/{dataset_id}/tags/'
    response = _get_request(dst_url, params=payload)
    tag_ids = [t['_id'] for t in response.json() if t['name'] == tag_name]
    if len(tag_ids) == 0:
        msg = f'No tag with name {tag_name} found '
        msg += f'for datset with id {dataset_id}'
        raise RuntimeError(msg)
    elif len(tag_ids) > 1:
        msg = f'{len(tag_ids)} tags with name {tag_name} founr '
        msg += f'for dataset with id {dataset_id}'
        raise RuntimeError(msg)
    tag_id = tag_ids[0]

    # get files in tag
    dst_url = f'{SERVER_LOCATION}/users/datasets/'
    dst_url += f'{dataset_id}/tags/{tag_id}/download'

    response = _get_request(dst_url, params=payload)
    return response.text.splitlines()



import numpy as np
import torch

import yaml


def init_seed(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def load_config_file(config_path):
    Loader = yaml.FullLoader
    with open(config_path, 'r') as config_file:
        cfg = yaml.load(config_file, Loader=Loader)
    return cfg


def count_samples_per_class(dataloader, num_classes):
    labels = []
    for batch, target, _ in dataloader:
        labels.extend(target.cpu().numpy())
    return [labels.count(i) for i in range(num_classes)]


def calculate_class_weights(dataloader, num_classes):
    n_samples_per_class = count_samples_per_class(dataloader, num_classes)
    max_samples_per_class = max(n_samples_per_class)
    weight = [max_samples_per_class / x for x in n_samples_per_class]
    return torch.FloatTensor(weight)


def confusion_matrix(output, target, n_classes):

    _, argmax = output.max(-1)
    cm = torch.zeros((n_classes, n_classes))
    for yhat, y in zip(argmax, target):
        cm[y, yhat] += 1

    return cm



import numpy as np
import torch

import unittest

from boris.transforms import GaussianBlur


class TestGaussianBlur(unittest.TestCase):

    def test_on_np_array(self):
        m, M = 0.1, 2.0
        for w in range(1, 100):
            for h in range(1, 100):
                gb = GaussianBlur(int(0.1 * w), min=m, max=M)
                sample = np.random.randn(w, h)
                gb(sample)

    def test_on_torch_tensor(self):
        m, M = 0.1, 2.0
        for w in range(1, 100):
            for h in range(1, 100):
                gb = GaussianBlur(int(0.1 * w), min=m, max=M)
                sample = torch.randn(w, h)
                gb(sample)



import unittest
import json
import os

import random
import responses
import numpy as np

import torchvision
import tempfile
from boris.api import upload_embeddings_from_csv
from boris.utils import embeddings_to_pandas


class TestUploadEmbeddings(unittest.TestCase):

    def setup(self, n_data=1000):

        # set up url
        self.dst_url = os.getenvb(b'BORIS_SERVER_LOCATION',
                                  b'https://api-dev.whattolabel.com').decode()
        self.dataset_id = 'XYZ'
        self.token = 'secret'

        # create a dataset
        self.dataset = torchvision.datasets.FakeData(size=n_data,
                                                     image_size=(3, 32, 32))

        self.folder_path = tempfile.mkdtemp()
        self.path_to_embeddings = os.path.join(
            self.folder_path,
            'embeddings.csv')

        sample_names = [f'img_{i}.jpg' for i in range(n_data)]
        labels = [0] * len(sample_names)

        df = embeddings_to_pandas(
            np.random.randn(n_data, 16),
            labels,
            sample_names
        )

        df.to_csv(self.path_to_embeddings)

    @responses.activate
    def test_upload_embeddings_all_success(self):
        """Make sure upload of embeddings works.

        """

        self.setup(n_data=10000)

        def post_embedding_callback(request, psuccess=1.):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/embeddings',
            callback=post_embedding_callback,
            content_type='application/json'
        )

        success = upload_embeddings_from_csv(
            self.path_to_embeddings,
            dataset_id=self.dataset_id,
            token=self.token
        )

        self.assertTrue(success)

    @responses.activate
    def test_upload_embeddings_random_success(self):
        """Make sure upload of embeddings works.

        """

        self.setup(n_data=5000)

        def post_embedding_callback(request, psuccess=.75):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/embeddings',
            callback=post_embedding_callback,
            content_type='application/json'
        )

        success = upload_embeddings_from_csv(
            self.path_to_embeddings,
            dataset_id=self.dataset_id,
            token=self.token
        )

        self.assertTrue(success)

    @responses.activate
    def test_upload_embeddings_no_success(self):
        """Make sure upload of embeddings works.

        """

        self.setup(n_data=5000)

        def post_embedding_callback(request, psuccess=.0):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/embeddings',
            callback=post_embedding_callback,
            content_type='application/json'
        )

        with self.assertRaises(RuntimeError):
            upload_embeddings_from_csv(
                self.path_to_embeddings,
                dataset_id=self.dataset_id,
                token=self.token
            )



""" boris.embedding

    The boris.embedding module provides trainable self-supervised
    embedding strategies.

"""

from boris import is_lightning_available

if is_lightning_available():
    from ._base import BaseEmbedding
    from ._embedding import SelfSupervisedEmbedding



import unittest

import numpy as np
import torch

from boris.sampling import sample


class TestSampler(unittest.TestCase):

    def create_state(self, n, d, device='cpu'):
        embeddings = torch.from_numpy(np.random.rand(n, d)).float().to(device)
        n_data = len(embeddings)
        selected = torch.from_numpy(np.zeros(n_data)).bool().to(device)
        state = (n_data, selected, embeddings, None)
        return state

    def test_all_samplers_cpu(self):
        STRATEGIES = ('bit', 'random', 'coreset')
        N_SAMPLES = (1e1, 1e2, 1e3)
        DIMENSION = (2, 16, 32)
        N_OUTPUT = int(1e1)

        for strategy in STRATEGIES:
            for n in N_SAMPLES:
                for d in DIMENSION:
                    n = int(n)
                    state = self.create_state(n, d, 'cpu')
                    new_state, sscores = sample(N_OUTPUT, state, strategy)

                    n_data, selected, embeddings, scores = new_state

                    self.assertIsInstance(n_data, int)
                    self.assertIsInstance(selected, torch.BoolTensor)
                    self.assertIsInstance(embeddings, torch.FloatTensor)
                    self.assertEqual(n_data, len(embeddings))
                    self.assertEqual(n_data, n)
                    self.assertTrue(torch.max(sscores).item() <= 1.0)
                    self.assertTrue(torch.min(sscores).item() >= 0.0)

    def test_bit_sampler_too_large_embedding(self):
        embeddings = torch.from_numpy(np.random.rand(10, 65)).float()
        n_data = len(embeddings)
        selected = torch.from_numpy(np.zeros(n_data)).bool()
        state = (n_data, selected, embeddings, None)

        with self.assertRaises(ValueError):
            new_state, sscores = sample(5, state, 'bit')



""" boris.models

    The boris.models module provides a collection of 
    models (such as ResNet) which can easily be used 
    for self-supervised learning.

"""

from ._resnet_simclr import ResNetSimCLR
from ._resnet_simclr import ResNetGenerator



import unittest
import json
import os

import random
import responses

import torchvision
import tempfile
import boris.api as api


class TestUploadImages(unittest.TestCase):

    def setup(self, n_data=1000):

        # set up url
        self.dst_url = os.getenvb(b'BORIS_SERVER_LOCATION',
                                  b'https://api-dev.whattolabel.com').decode()
        self.dataset_id = 'XYZ'
        self.token = 'secret'

        # create a dataset
        self.dataset = torchvision.datasets.FakeData(size=n_data,
                                                     image_size=(3, 32, 32))

        self.folder_path = tempfile.mkdtemp()
        sample_names = [f'img_{i}.jpg' for i in range(n_data)]
        for sample_idx in range(n_data):
            data = self.dataset[sample_idx]
            path = os.path.join(self.folder_path, sample_names[sample_idx])
            data[0].save(path)

    @responses.activate
    def test_upload_images_all_success(self):
        """Make sure upload works.

        """

        self.setup(n_data=50)

        def signed_url_callback(request, psuccess=1.):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        def put_image_callback(request, psuccess=1.):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        responses.add_callback(
            responses.PUT, 'https://this-is-a-url.com',
            callback=put_image_callback,
            content_type='application/json'
        )

        api.upload_images_from_folder(
            self.folder_path,
            dataset_id=self.dataset_id,
            token=self.token
        )

    @responses.activate
    def test_upload_images_random_success(self):
        """Make sure test runs through eventhough not all files get uploaded.

        """

        self.setup(n_data=50)

        def signed_url_callback(request, psuccess=.7):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        def put_image_callback(request, psuccess=.7):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        responses.add_callback(
            responses.PUT, 'https://this-is-a-url.com',
            callback=put_image_callback,
            content_type='application/json'
        )

        api.upload_images_from_folder(
            self.folder_path,
            dataset_id=self.dataset_id,
            token=self.token
        )

    @responses.activate
    def test_upload_images_no_success(self):
        """Make sure test runs through eventhough not all files get uploaded.

        """

        self.setup(n_data=50)

        def signed_url_callback(request, psuccess=1.):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        def put_image_callback(request, psuccess=.0):
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps({}))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        responses.add_callback(
            responses.PUT, 'https://this-is-a-url.com',
            callback=put_image_callback,
            content_type='application/json'
        )

        with self.assertWarns(Warning):
            api.upload_images_from_folder(
                self.folder_path,
                dataset_id=self.dataset_id,
                token=self.token
            )



""" Embeddings """

import time

import torch
from boris import is_prefetch_generator_available
from boris.embedding._base import BaseEmbedding
from tqdm import tqdm

if is_prefetch_generator_available():
    from prefetch_generator import BackgroundGenerator


class SelfSupervisedEmbedding(BaseEmbedding):
    """ Self-supervised embedding based on contrastive multiview coding.
        (https://arxiv.org/abs/1906.05849)

    """

    def __init__(self, model, criterion, optimizer, dataloader):
        """ Constructor

        Args:
            model: (torch.nn.Module)
            criterion: (torch.nn.Module)
            optimizer: (torch.optim.Optimizer)
            dataloader: (torch.utils.data.DataLoader)

        """

        super(SelfSupervisedEmbedding, self).__init__(
            model, criterion, optimizer, dataloader
        )

    def embed(self, dataloader, caching=False, normalize=False, device=None):
        """ Embed data in vector space

        Args:
            dataloader: (torch.utils.data.DataLoader)
            caching: (bool) TODO
            normalize: (bool) Normalize embeddings to unit length

        """

        # TODO: Caching
        if caching:
            pass
            return None, None

        self.model.eval()
        embeddings, labels, fnames = None, None, []

        if is_prefetch_generator_available():
            pbar = tqdm(BackgroundGenerator(dataloader, max_prefetch=3),
                        total=len(dataloader))
        else:
            pbar = tqdm(dataloader, total=len(dataloader))

        start_time = time.time()
        with torch.no_grad():

            for (img, label, fname) in pbar:

                img = img.to(device)
                label = label.to(device)

                fnames += [*fname]

                batch_size = img.shape[0]
                prepare_time = time.time()

                emb = self.model.features(img)
                emb = emb.detach().reshape(batch_size, -1)

                if embeddings is None:
                    embeddings = emb
                else:
                    embeddings = torch.cat((embeddings, emb), 0)

                if labels is None:
                    labels = label
                else:
                    labels = torch.cat((labels, label), 0)

                process_time = time.time()

                pbar.set_description("Compute efficiency: {:.2f}".format(
                    process_time / (process_time + prepare_time)))

            embeddings = embeddings.cpu().numpy()
            labels = labels.cpu().numpy()

        return embeddings, labels, fnames


class VAEEmbedding(BaseEmbedding):
    """ Unsupervised embedding based on variational auto-encoders.

    """

    def embed(self, dataloader):
        """ TODO

        """
        raise NotImplementedError("This site is under construction...")



import unittest

import torch

from boris.models import ResNetSimCLR


class TestModelsSimCLR(unittest.TestCase):

    def setUp(self):
        self.resnet_variants = [
            'resnet-18',
            'resnet-34',
            'resnet-50',
            'resnet-101',
            'resnet-152'
        ]
        self.batch_size = 4
        self.input_tensor = torch.rand((self.batch_size, 3, 32, 32))

    def test_create_variations_cpu(self):
        for model_name in self.resnet_variants:
            model = ResNetSimCLR(name=model_name)
            self.assertIsNotNone(model)

    def test_create_variations_gpu(self):
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        if device == 'cuda':
            for model_name in self.resnet_variants:
                model = ResNetSimCLR(name=model_name).to(device)
                self.assertIsNotNone(model)
        else:
            pass

    def test_feature_dim_configurable(self):
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        for model_name in self.resnet_variants:
            for num_ftrs in [2, 16, 32, 64]:
                for out_dim in [64, 128, 256]:
                    model = ResNetSimCLR(name=model_name,
                                         num_ftrs=num_ftrs,
                                         out_dim=out_dim).to(device)

                    # check that feature vector has correct dimension
                    with torch.no_grad():
                        out_features = model.features(
                            self.input_tensor.to(device)
                        )
                    self.assertEqual(out_features.shape[1], num_ftrs)

                    # check that projection head output has right dimension
                    with torch.no_grad():
                        out_projection = model.projection_head(
                            out_features.squeeze()
                        )
                    self.assertEqual(out_projection.shape[1], out_dim)
                    self.assertIsNotNone(model)

    def test_variations_input_dimension(self):
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        for model_name in self.resnet_variants:
            for input_width in [16, 32, 64, 256]:
                for input_height in [16, 32, 64, 256]:
                    model = ResNetSimCLR(name=model_name).to(device)

                    input_tensor = torch.rand((self.batch_size,
                                               3,
                                               input_height,
                                               input_width))
                    with torch.no_grad():
                        out = model(input_tensor.to(device))

                    self.assertIsNotNone(model)
                    self.assertIsNotNone(out)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    boris-cli
    ~~~~~~~~~

    A simple command line tool to train a state-of-the-art
    self-supervised model on your image dataset.

"""

import hydra
import warnings

from boris.data import ImageCollateFunction
from boris.data import BorisDataset
from boris.embedding import SelfSupervisedEmbedding
from boris.loss import NTXentLoss
from boris.models import ResNetSimCLR

from boris.cli._helpers import _is_url
from boris.cli._helpers import _model_zoo
from boris.cli._helpers import fix_input_path
from boris.cli._helpers import filter_state_dict

import torch


def _train_cli(cfg, is_cli_call=True):
    """Train a self-supervised model on the image dataset of your choice.

    Args:
        cfg[data]: (str)
            Name of the dataset (to download use cifar10 or cifar100)
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str)
            If specified, the dataset is loaded from the folder

    Returns:
        checkpoint: (str) Path to checkpoint of the best model during training

    """

    data = cfg['data']
    download = cfg['download']

    root = cfg['root']
    if root and is_cli_call:
        root = fix_input_path(root)

    from_folder = cfg['from_folder']
    if from_folder and is_cli_call:
        from_folder = fix_input_path(from_folder)

    if 'seed' in cfg.keys():
        seed = cfg['seed']
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    if torch.cuda.is_available():
        device = 'cuda'
    elif cfg['trainer'] and cfg['trainer']['gpus']:
        device = 'cpu'
        cfg['trainer']['gpus'] = 0

    if cfg['loader']['batch_size'] < 64:
        msg = 'Training a self-supervised model with a small batch size: {}! '
        msg = msg.format(cfg['loader']['batch_size'])
        msg += 'Small batch size may harm embedding quality. '
        msg += 'You can specify the batch size via the loader key-word: '
        msg += 'loader.batch_size=BSZ'
        warnings.warn(msg)

    model = ResNetSimCLR(**cfg['model'])
    checkpoint = cfg['checkpoint']
    if cfg['pre_trained'] and not checkpoint:
        # if checkpoint wasn't specified explicitly and pre_trained is True
        # try to load the checkpoint from the model zoo
        checkpoint, key = _model_zoo(cfg['model'])
        if not checkpoint:
            msg = 'Cannot download checkpoint for key {} '.format(key)
            msg += 'because it does not exist! '
            msg += 'Model will be trained from scratch.'
            warnings.warn(msg)
    elif checkpoint:
        checkpoint = fix_input_path(checkpoint) if is_cli_call else checkpoint
    
    if checkpoint:
        # load the PyTorch state dictionary and map it to the current device
        # then, remove the model. prefix which is caused by the pytorch-lightning
        # checkpoint saver and load the model from the "filtered" state dict
        # this approach is compatible with pytorch_lightning 0.7.1 - 0.8.4 (latest)
        if _is_url(checkpoint):
            state_dict = torch.hub.load_state_dict_from_url(
                checkpoint, map_location=device
            )['state_dict']
        else:
            state_dict = torch.load(
                checkpoint, map_location=device
            )['state_dict']
        state_dict = filter_state_dict(state_dict)
        model.load_state_dict(state_dict)
        
    criterion = NTXentLoss(**cfg['criterion'])
    optimizer = torch.optim.SGD(model.parameters(), **cfg['optimizer'])

    dataset = BorisDataset(root,
                           name=data, train=True, download=download,
                           from_folder=from_folder)

    collate_fn = ImageCollateFunction(**cfg['collate'])
    dataloader = torch.utils.data.DataLoader(dataset,
                                             **cfg['loader'],
                                             collate_fn=collate_fn)

    encoder = SelfSupervisedEmbedding(model, criterion, optimizer, dataloader)
    encoder = encoder.train_embedding(**cfg['trainer'])

    print('Best model is stored at: %s' % (encoder.checkpoint))
    return encoder.checkpoint


@hydra.main(config_path="config/config.yaml", strict=False)
def train_cli(cfg):
    """Train a self-supervised model on the image dataset of your choice.

    Args:
        cfg[data]: (str)
            Name of the dataset (to download use cifar10 or cifar100)
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str)
            If specified, the dataset is loaded from the folder

    Returns:
        checkpoint: (str) Path to checkpoint of the best model during training

    """

    return _train_cli(cfg)


def entry():
    train_cli()



import torch

from boris.sampling._score import _inf
from boris.sampling._score import _random_score_fn
from boris.sampling._score import _coreset_score_fn
from boris.sampling._score import _bit_score_fn
from boris.sampling._score import _uncertainty_score_fn


def _permutation_sample(scores):
    _, indices = torch.sort(scores, descending=True)
    return indices


def _unpack_state(state):
    """Unpack state into 4-tuple.

    """

    if type(state) is tuple:
        if len(state) == 4:
            return state

    elif type(state) is torch.Tensor:
        n_data = state.shape[0]
        selected = torch.zeros(n_data).bool().to(state.device)
        return (n_data, selected, state, None)

    length = len(state) if type(state) is tuple else 0
    raise ValueError(
        'Illegal state! Must be 4-tuple or torch.Tensor but is {} with len {}'
        .format(type(state), length)
    )


def _parse_state(state):
    """Parse state and make sure format is ok.

    """

    n_data, selected, embeddings, uscores = _unpack_state(state)

    if selected.dtype is not torch.bool:
        raise ValueError(
            'selected must be dtype torch.bool, is {}.'
            .format(selected.dtype)
        )

    if embeddings is not None and embeddings.dtype is not torch.float:
        raise ValueError(
            'embeddings must be dtype torch.float, is {}'
            .format(embeddings.dtype)
        )

    if uscores is not None and uscores.dtype is not torch.float:
        raise ValueError(
            'uscores must be dtype torch.float, is {}'
            .format(uscores.dtype)
        )

    return (n_data, selected, embeddings, uscores)


def _sample(n_samples, state, score_fn=_random_score_fn, **kwargs):
    """ Update sample state based on current state, embeddings,
        and the score function.

    Args:
        n_samples: (int) Number of samples
        state: (n_data, selected, embeddings, scores)
            n_data: (int) size of the dataset
            selected: (torch.BoolTensor) selected samples
            embeddings: (torch.FloatTensor) TODO
            uscores: (torch.FloatTensor) uncertainty scores
        score_fn: (callable) computes new importance scores
            based on current scores and embeddings

    Returns:
        new_state: (n_data, selected, embeddings, uscores)
        sscores: (torch.FloatTensor) sampling scores

    """

    n_data, selected, embeddings, uscores = state
    n_selected = selected.sum()

    if n_samples <= 0:
        return state
    if n_selected >= n_data:
        return state

    if embeddings is None and uscores is None:
        sscores = _random_score_fn(state)
    elif embeddings is not None:
        sscores = score_fn(state, **kwargs)

    max_val = _inf(sscores.max())
    min_val = sscores.min()
    sscores[selected] = max_val
    sscores = (sscores - min_val) / _inf(max_val - min_val)

    indices = _permutation_sample(sscores)

    n_samples = min(n_data, n_selected + n_samples)
    selected[indices[n_selected:n_samples]] = True

    return (n_data, selected, embeddings, uscores), sscores


def sample(n_samples, state, strategy='random', **kwargs):
    """ Update sample state based on current state, embeddings,
        and the chosen strategy.

    Args:
        n_samples: (int) Number of samples
        state: (n_data, selected, embeddings, scores)
            n_data: (int) size of the dataset
            selected: (torch.BoolTensor) selected samples
            embeddings: (torch.FloatTensor) TODO
            uscores: (torch.FloatTensor) uncertainty scores
        strategy: (str) sampling strategy from
            {random, coreset}

    Returns:
        new_state: (n_data, selected, embeddings, scores)
        sscores: (torch.FloatTensor) sampling scores

    Examples:

        >>> n_data = len(dataset)
        >>> selected = torch.zeros(n_data).bool()
        >>> state = (n_data, selected, embeddings, None)
        >>> state, sscores = sample(n_samples, state, strategy='coreset')

        >>> n_data = len(dataset)
        >>> selected = torch.zeros(n_data).bool()
        >>> selected[0] = True ## manual initial choice
        >>> uscores = # get uncertainty scores for active learning
        >>> state = (n_data, selected, embeddings, uscores)
        >>> state, sscores = sample(n_samples, state, strategy='coreset')
    """

    state = _parse_state(state)

    if strategy == 'random':
        score_fn = _random_score_fn
    elif strategy == 'coreset':
        score_fn = _coreset_score_fn
    elif strategy == 'bit':
        score_fn = _bit_score_fn
    elif strategy == 'uncertainty':
        score_fn = _uncertainty_score_fn
    else:
        raise ValueError(
            'Illegal strategy: {}'.format(strategy))

    state, sscores = _sample(n_samples, state, score_fn=score_fn, **kwargs)
    return state, sscores



""" Deep Learning Package for Python

boris is a Python module for self-supervised active learning.

"""

try:
    import pytorch_lightning
except ImportError:
    _lightning_available = False
else:
    _lightning_available = True

try:
    import cv2
except ImportError:
    _opencv_available = False
else:
    _opencv_available = True

try:
    import prefetch_generator
except ImportError:
    _prefetch_generator_available = False
else:
    _prefetch_generator_available = True


def is_lightning_available():
    return _lightning_available


def is_opencv_available():
    return _opencv_available


def is_prefetch_generator_available():
    return _prefetch_generator_available


if is_lightning_available():
    from ._one_liners import train_model_and_get_image_features
    from ._one_liners import train_self_supervised_model
    from ._one_liners import get_image_features

__version__ = '0.1.4'



# -*- coding: utf-8 -*-
"""
    boris.cli
    ~~~~~~~~~

    A simple command line tool to upload your datasets to
    the WhatToLabel platform.

"""

import hydra

from boris.api import upload_embeddings_from_csv
from boris.api import upload_images_from_folder

from boris.cli._helpers import fix_input_path


def _upload_cli(cfg, is_cli_call=True):
    '''Upload your image dataset and/or embeddings to the WhatToLabel platform.

    Args:
        cfg['path_to_folder']: (str)
            Path to folder which holds images to upload
        cfg['path_to_embeddings']: (str)
            Path to csv file which holds embeddings to upload
        cfg['dataset_id']: (str) Dataset identifier on the platform
        cfg['token']: (str) Token which grants acces to the platform

    '''

    path_to_folder = cfg['path_to_folder']
    if path_to_folder and is_cli_call:
        path_to_folder = fix_input_path(cfg['path_to_folder'])

    path_to_embeddings = cfg['path_to_embeddings']
    if path_to_embeddings and is_cli_call:
        path_to_embeddings = fix_input_path(cfg['path_to_embeddings'])

    dataset_id = cfg['dataset_id']
    token = cfg['token']

    if not token:
        raise ValueError('token was not specified!')
    if not dataset_id:
        raise ValueError('dataset_id was not specified!')

    if path_to_folder:
        upload_images_from_folder(path_to_folder, dataset_id, token)

    if path_to_embeddings:
        upload_embeddings_from_csv(path_to_embeddings, dataset_id, token)


@hydra.main(config_path='config/config.yaml')
def upload_cli(cfg):
    '''Upload your image dataset and/or embeddings to the WhatToLabel platform.

    Args:
        cfg['path_to_folder']: (str)
            Path to folder which holds images to upload
        cfg['path_to_embeddings']: (str)
            Path to csv file which holds embeddings to upload
        cfg['dataset_id']: (str) Dataset identifier on the platform
        cfg['token']: (str) Token which grants acces to the platform

    '''
    _upload_cli(cfg)


def entry():
    upload_cli()



""" boris.data

    The boris.data module provides a dataset wrapper
    and useful collate functions for self-supervised learning
    (see torch.utils.data.DataLoader for details on collate_fn)

"""

from ._dataset import BorisDataset
from ._collate import BaseCollateFunction
from ._collate import ImageCollateFunction



import numpy as np
import torch
import sys
import os

import hydra

import boris.sampling as sampling

from configure import configure_embedding
from configure import configure_evaluatable
from configure import configure_trainable
from helpers import init_seed
from helpers import load_config_file
from log import Logger
from log import dump_embeddings

HYDRA_PATH_PREFIX = '../../../config/'


@hydra.main(config_path='config/config.yaml', strict=False)
def main(cfg):
    """
        BENCHMARK TIME COMPLEXITY:
        the benchmark trains and evaluates:

            #repetitions x #embedding_configs x #sample_configs x #percentages
            -> this takes a LOT of time!
    """

    # redirect output
    out = open('benchmark.log.txt', 'w')
    sys.stdout = out

    # benchmark every ten percent of the dataset
    sample_percentages = np.arange(.25, 1.1, .25)

    # check for cuda
    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    # if a sampler_cfg file was passed -> use it, otherwise load all
    sampler_cfg = []
    if cfg['sampler_cfg']:
        path_to_config = os.path.join(HYDRA_PATH_PREFIX, 'sampler/')
        path_to_config = os.path.join(path_to_config, cfg['sampler_cfg'])
        sampler_cfg = [path_to_config]
    else:
        path_to_config = os.path.join(HYDRA_PATH_PREFIX, 'sampler/')
        files = os.listdir(path_to_config)
        sampler_cfg = [path_to_config + cfg for cfg in files]

    # if an embedding_cfg file was passed -> use it, otherwise load all
    embedding_cfg = []
    if cfg['embedding_cfg']:
        path_to_config = os.path.join(HYDRA_PATH_PREFIX, 'embedding/')
        path_to_config = os.path.join(path_to_config, cfg['embedding_cfg'])
        embedding_cfg = [path_to_config]
    else:
        path_to_config = os.path.join(HYDRA_PATH_PREFIX, 'embedding/')
        files = os.listdir(path_to_config)
        embedding_cfg = [path_to_config + cfg for cfg in files]

    # initialize the logger to keep track of results
    log = Logger(cfg['name'])

    # run benchmark n_repetitions times for each setting
    for rep in range(cfg['n_repetitions']):

        # fix seed for reproducibility
        init_seed(rep)

        # run benchmark for each embedding configuration
        for i in range(len(embedding_cfg)):

            try:

                # get new embedding model
                embedding_cfg_inst = load_config_file(embedding_cfg[i])

                embedding_model, embedding_loader = \
                    configure_embedding(embedding_cfg_inst,
                                        cfg['training']['data'],
                                        cfg['training']['transform'])

                # train embedding and get embeddings
                embedding_model = embedding_model.train_embedding(
                    **embedding_cfg_inst['trainer'])

                embeddings, labels, _ = embedding_model.embed(
                    embedding_loader, device=device)

                dump_embeddings(embeddings, labels, rep)

                embeddings = torch.from_numpy(
                    embeddings.astype(float)
                    ).float().to(device)

                # run benchmark for each sampler configuration
                for j in range(len(sampler_cfg)):

                    # load sample cfg file
                    sampler_cfg_inst = load_config_file(sampler_cfg[j])

                    # make new sampling state
                    n_data = embeddings.shape[0]
                    selected = torch.zeros(n_data).bool().to(device)
                    state = (n_data, selected, embeddings, None)

                    n_samples = int(sample_percentages[0] * n_data)

                    # run benchmark for each sample percentage
                    for k in range(len(sample_percentages)):

                        # sample
                        print(sampler_cfg_inst)
                        state, _ = sampling.sample(n_samples, state,
                                                   **sampler_cfg_inst)

                        # train
                        trainable, trainer = configure_trainable(cfg, state)
                        trainer.fit(trainable)

                        # evaluate
                        evaluatable = configure_evaluatable(cfg)
                        evaluatable.evaluate(trainable, device=device)

                        # make a backup of the results
                        row = [rep]
                        row += [os.path.basename(embedding_cfg[i])]
                        row += [os.path.basename(sampler_cfg[j])]
                        row += [sample_percentages[k]]
                        row += [
                            metric.result
                            for metric in evaluatable.eval_metrics
                        ]
                        log.add_row(row)

                        header = ['run_id']
                        header += ['embedding_model']
                        header += ['sampling_strategy']
                        header += ['sample_percentage']
                        header += [
                            metric.identifier
                            for metric in evaluatable.eval_metrics
                        ]
                        log.backup(header)

                        if not cfg['active_learning']:
                            # make new sampling state
                            selected = torch.zeros(n_data).bool().to(device)
                            state = (n_data, selected, embeddings, None)
                            if k + 1 < len(sample_percentages):
                                n_samples = sample_percentages[k+1] * n_data
                                n_samples = int(n_samples)
                        else:
                            state = (n_data, selected, embeddings, None)
                            n_samples = int(sample_percentages[0] * n_data)

            except KeyError:
                print('WARNING: Benchmark encountered: KeyError!')
                print('Skipping iteration...')
            except NotImplementedError:
                print('WARNING: Benchmark encountered: NotImplementedError!')
                print('Skipping iteration...')
            except ValueError:
                print('WARNING: Benchmark encountered: ValueError!')
                print('Skipping iteration...')


if __name__ == '__main__':
    main()



import unittest
import torch

from boris.loss import NTXentLoss


class TestNTXentLoss(unittest.TestCase):

    def test_get_correlated_mask(self):
        loss = NTXentLoss()
        for bsz in range(1, 1000):
            mask = loss._torch_get_correlated_mask(bsz)

            # correct number of zeros in mask
            self.assertAlmostEqual(mask.sum(), 4*(bsz*bsz - bsz))

            # if mask is correct,
            # (1 - mask) * v adds up the first and second half of v
            v = torch.randn((2*bsz))
            mv = torch.mv(1. - mask.float(), v)
            vv = (v[bsz:] + v[:bsz]).repeat(2)
            self.assertAlmostEqual((mv - vv).pow(2).sum(), 0.)

    def test_forward_pass(self):
        loss = NTXentLoss()
        for bsz in range(1, 100):

            batch_1 = torch.randn((bsz, 32))
            batch_2 = torch.randn((bsz, 32))

            # symmetry
            l1 = loss(torch.cat((batch_1, batch_2), 0))
            l2 = loss(torch.cat((batch_2, batch_1), 0))
            self.assertAlmostEqual((l1 - l2).pow(2).item(), 0.)

    def test_forward_pass_1d(self):
        loss = NTXentLoss()
        for bsz in range(1, 100):

            batch_1 = torch.randn((bsz, 1))
            batch_2 = torch.randn((bsz, 1))

            # symmetry
            l1 = loss(torch.cat((batch_1, batch_2), 0))
            l2 = loss(torch.cat((batch_2, batch_1), 0))
            self.assertAlmostEqual((l1 - l2).pow(2).item(), 0.)

    def test_forward_pass_neg_temp(self):
        loss = NTXentLoss(temperature=-1.)
        for bsz in range(1, 100):

            batch_1 = torch.randn((bsz, 32))
            batch_2 = torch.randn((bsz, 32))

            # symmetry
            l1 = loss(torch.cat((batch_1, batch_2), 0))
            l2 = loss(torch.cat((batch_2, batch_1), 0))
            self.assertAlmostEqual((l1 - l2).pow(2).item(), 0.)

    def test_forward_pass_cuda(self):
        if torch.cuda.is_available():
            loss = NTXentLoss()
            for bsz in range(1, 100):

                batch_1 = torch.randn((bsz, 32)).cuda()
                batch_2 = torch.randn((bsz, 32)).cuda()

                # symmetry
                l1 = loss(torch.cat((batch_1, batch_2), 0))
                l2 = loss(torch.cat((batch_2, batch_1), 0))
                self.assertAlmostEqual((l1 - l2).pow(2).item(), 0.)
        else:
            pass



import unittest
import json
import os

import random
import responses

import torchvision
from boris.api import get_presigned_upload_url


class TestGetPresignedURL(unittest.TestCase):

    def setup(self, n_data=1000):

        # set up url
        self.dst_url = os.getenvb(b'BORIS_SERVER_LOCATION',
                                  b'https://api-dev.whattolabel.com').decode()
        self.dataset_id = 'XYZ'
        self.token = 'secret'

        # create a dataset
        self.dataset = torchvision.datasets.FakeData(size=n_data,
                                                     image_size=(3, 32, 32))

    @responses.activate
    def test_signed_url_all_success(self):
        '''Make sure everything works in no-error scenario.

        '''

        self.setup(n_data=10)

        def signed_url_callback(request, psuccess=1.):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        for f in self.dataset:
            url, success = get_presigned_upload_url(
                'filename',
                self.dataset_id,
                self.token
            )

    @responses.activate
    def test_signed_url_random_success(self):
        '''Make sure exponential backoff works and we get the url at the end.

        '''

        self.setup(n_data=10)

        def signed_url_callback(request, psuccess=.75):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        for f in self.dataset:
            url, success = get_presigned_upload_url(
                'filename',
                self.dataset_id,
                self.token
            )

    @responses.activate
    def test_signed_url_no_success(self):
        """Make sure we get a RuntimeError when we do not receive a url back

        """

        self.setup(n_data=0)

        def signed_url_callback(request, psuccess=0.):
            resp_body = {'url': 'https://this-is-a-url.com'}
            headers = {'request-id': '728d329e-0e86-11e4-a748-0c84dc037c13'}
            if random.random() < psuccess:
                return (200,
                        headers,
                        json.dumps(resp_body))
            else:
                return (500,
                        headers,
                        json.dumps({}))

        responses.add_callback(
            responses.POST, f'{self.dst_url}/getsignedurl',
            callback=signed_url_callback,
            content_type='application/json'
        )

        with self.assertRaises(RuntimeError):
            url, success = get_presigned_upload_url(
                'filename',
                self.dataset_id,
                self.token
            )



# -*- coding: utf-8 -*-
"""
    boris.cli
    ~~~~~~~~~

    A simple command line tool to embed your image dataset using
    a pretrained state-of-the-art, self-supervised model.

"""

import os

import hydra
import torch
import torchvision

from boris.data import BorisDataset
from boris.embedding import SelfSupervisedEmbedding
from boris.models import ResNetSimCLR
from boris.utils import embeddings_to_pandas

from boris.cli._helpers import _is_url
from boris.cli._helpers import _model_zoo
from boris.cli._helpers import fix_input_path
from boris.cli._helpers import filter_state_dict


def _embed_cli(cfg, is_cli_call=True):
    """Use the trained self-supervised model to embed samples from your dataset.

    Args:
        cfg[data]: (str) Name of the dataset
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[checkpoint]: (str) Path to the lightning checkpoint
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str) If specified, the dataset is loaded \
            from the folder

    Returns:
        embeddings: (np.ndarray) A d-dimensional embedding \
            for each data sample
        labels: (np.ndarray) Data labels, 0 if there are no labels
        filenames: (List[str]) File name of each data sample
    """

    data = cfg['data']
    checkpoint = cfg['checkpoint']
    download = cfg['download']

    root = cfg['root']
    if root and is_cli_call:
        root = fix_input_path(root)

    from_folder = cfg['from_folder']
    if from_folder and is_cli_call:
        from_folder = fix_input_path(from_folder)

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    model = ResNetSimCLR(**cfg['model']).to(device)

    transform = torchvision.transforms.Compose([
        torchvision.transforms.Resize((cfg['collate']['input_size'],
                                       cfg['collate']['input_size'])),
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225])
    ])

    dataset = BorisDataset(root,
                           name=data, train=True, download=download,
                           from_folder=from_folder, transform=transform)

    cfg['loader']['drop_last'] = False
    cfg['loader']['shuffle'] = False
    dataloader = torch.utils.data.DataLoader(dataset, **cfg['loader'])

    # load the PyTorch state dictionary and map it to the current device
    # then, remove the model. prefix which is caused by the pytorch-lightning
    # checkpoint saver and load the model from the "filtered" state dict
    # this approach is compatible with pytorch_lightning 0.7.1 - 0.8.4 (latest)
    if not checkpoint:
        checkpoint, key = _model_zoo(cfg['model'])
        if not checkpoint:
            msg = 'Cannot download checkpoint for key {} '.format(key)
            msg += 'because it does not exist!'
            raise RuntimeError(msg)
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=device
        )['state_dict']
    else:
        checkpoint = fix_input_path(checkpoint) if is_cli_call else checkpoint
        state_dict = torch.load(
            checkpoint, map_location=device
        )['state_dict']
    state_dict = filter_state_dict(state_dict)
    model.load_state_dict(state_dict)
    encoder = SelfSupervisedEmbedding(model, None, None, None)

    embeddings, labels, filenames = encoder.embed(dataloader, device=device)

    df = embeddings_to_pandas(embeddings, labels, filenames)
    path = os.path.join(os.getcwd(), 'embeddings.csv')
    df.to_csv(path, index=False)
    print('Embeddings are stored at %s' % (path))

    return embeddings, labels, filenames


@hydra.main(config_path='config/config.yaml', strict=False)
def embed_cli(cfg):
    """Use the trained self-supervised model to embed samples from your dataset.

    Args:
        cfg[data]: (str) Name of the dataset
        cfg[root]: (str) Directory where the dataset should be stored
        cfg[checkpoint]: (str) Path to the lightning checkpoint
        cfg[download]: (bool) Whether to download the dataset
        cfg[from_folder]: (str) If specified, the dataset is loaded \
            from the folder

    Returns:
        embeddings: (np.ndarray) A d-dimensional embedding \
            for each data sample
        labels: (np.ndarray) Data labels, 0 if there are no labels
        filenames: (List[str]) File name of each data sample
    """

    return _embed_cli(cfg)


def entry():
    embed_cli()



# -*- coding: utf-8 -*-
"""
    boris.cli
    ~~~~~~~~~

    A simple command line tool to embed your image dataset with
    state-of-the-art technology.

"""

import hydra

from boris.cli.train_cli import _train_cli
from boris.cli.embed_cli import _embed_cli


def _boris_cli(cfg, is_cli_call=True):
    """Train a self-supervised model and use it to embed your dataset

    Args:
        see train_cli.py, embed_cli.py or
        https://www.notion.so/whattolabel/WhatToLabel-Documentation-28e645f5564a453e807d0a384a4e6ea7

    """

    cfg['loader']['shuffle'] = True
    cfg['loader']['drop_last'] = True
    checkpoint = _train_cli(cfg, is_cli_call)

    cfg['loader']['shuffle'] = False
    cfg['loader']['drop_last'] = False
    cfg['checkpoint'] = checkpoint

    return _embed_cli(cfg, is_cli_call)


@hydra.main(config_path="config/config.yaml", strict=False)
def boris_cli(cfg):
    """Train a self-supervised model and use it to embed your dataset

    Args:
        see train_cli.py, embed_cli.py or
        https://www.notion.so/whattolabel/WhatToLabel-Documentation-28e645f5564a453e807d0a384a4e6ea7

    """
    return _boris_cli(cfg)


def entry():
    boris_cli()



import torch
import numpy as np


def _inf(M=1, eps=1e-8):
    return M + eps


def _pairwise_distance(X, Y):
    X_norm = (X ** 2).sum(1).view(-1, 1)
    Y_norm = (Y ** 2).sum(1).view(1, -1)
    XY = torch.mm(X, torch.transpose(Y, 0, 1))
    return X_norm + Y_norm - 2*XY


def _random_score_fn(state):
    """ random sampling score

    """
    n_data, selected, embeddings, scores = state
    return torch.randn_like(selected.float())


def _coreset_score_fn(state):
    """ coreset sampling score

    """

    n_data, selected, embeddings, uscores = state

    selected_clone = selected.clone()
    sscores = torch.zeros_like(selected).float()

    choice, nn_dist = None, None
    if not selected_clone.any():
        # make initial selection
        choice = torch.randint(0, n_data, [1]).to(selected.device)
        selected_clone[choice] = True
        nn_dist = torch.norm(embeddings - embeddings[choice], dim=1).pow(2)
        nn_dist_tmp = nn_dist if uscores is None else nn_dist * uscores
        sscores[choice] = nn_dist_tmp.max()

    n_selected = selected_clone.sum()
    while n_selected < n_data:

        batch = embeddings[~selected_clone]

        if choice is None:
            nn_dist = torch.zeros_like(selected).float()
            nn_dist[~selected_clone], _ = _pairwise_distance(
                batch, embeddings[selected_clone]
                ).min(1)
        else:
            dist = torch.norm(batch - embeddings[choice], dim=1).pow(2)
            nn_dist[~selected_clone] = torch.min(
                nn_dist[~selected_clone], dist)

        nn_dist_tmp = nn_dist if uscores is None else nn_dist * uscores

        sscore, choice = nn_dist_tmp.max(-1)
        selected_clone[choice] = True
        nn_dist[choice] = -_inf()
        sscores[choice] = sscore
        n_selected += 1

    sscores[selected] = _inf(sscores.max())
    return sscores


def _bit_score_fn(state, depth=None):
    """ quick but memory intense

    """

    n_data, selected, embeddings, scores = state
    _, n_dim = embeddings.shape

    if depth is None:
        # n_data ~= C * 2 ** (n_dim * depth)
        depth = int(np.log2(n_data)) - n_dim
        depth = max(min(8, depth), 1)

    if depth > 8:
        raise ValueError('Illegal depth (depth > 8): {}'.format(depth))

    if depth * n_dim > 62 or depth * n_dim <= 0:
        raise ValueError(
            'Illegal combination of depth {} and number of dimensions {}! \
            Requires 0 < depth * n_dim <= 64'
            .format(depth, n_dim))

    bitreps = torch.zeros_like(selected).long()

    m, _ = embeddings.min(0)
    M, _ = embeddings.max(0)

    intreps = (embeddings - m - 1e-5) / (M - m)
    intreps[intreps < 0.] = 0.
    intreps *= 2 ** depth
    intreps = intreps.long()

    bitreps = torch.zeros_like(selected).long()
    for d in range(n_dim):
        exp = (n_dim - d - 1) * depth
        shift = 2 ** exp
        bitreps += intreps[:, d] * shift

    sscores = torch.zeros_like(selected).float()
    lookup = {}

    # add previously selected to lookup
    sscores[selected] = _inf()
    indices = selected.nonzero()
    for i in indices:
        bitrep = bitreps[i].item()
        lookup[bitrep] = 1

    # sample next set
    indices = (~selected).nonzero()
    for i in indices[torch.randperm(len(indices))]:
        bitrep = bitreps[i].item()
        if lookup.get(bitrep) is None:
            lookup[bitrep] = 1
            sscores[i] = 1.
        else:
            lookup[bitrep] += 1
            sscores[i] = 1. / lookup[bitrep]

    # add tiny noise
    sscores = torch.normal(sscores,
                           1e-6 * torch.ones(n_data).to(selected.device))

    return sscores


def _uncertainty_score_fn(state):
    """Returns the uncertainty score for each sample.
       Useful for active learning without diversity sampling.

    """
    _, selected, _, scores = state
    if scores is None:
        msg = 'Uncertainty scores in state must not be None! '
        msg += 'Received state (_, _, _, uscore) = {}'.format(state)
        raise ValueError(msg)

    return scores



""" boris.api

    The boris.api module provides access to the WhatToLabel web-app.
"""

from ._communication import get_presigned_upload_url
from ._helpers import upload_images_from_folder
from ._helpers import upload_file_with_signed_url
from ._helpers import upload_embeddings_from_csv
from ._helpers import get_samples_by_tag



"""torchvision image loaders

(see https://pytorch.org/docs/stable/_modules/torchvision/datasets/folder.html)
"""
from PIL import Image


def pil_loader(path):
    # open path as file to avoid ResourceWarning
    # (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert('RGB')


def accimage_loader(path):
    import accimage
    try:
        return accimage.Image(path)
    except IOError:
        # Potentially a decoding problem, fall back to PIL.Image
        return pil_loader(path)


def default_loader(path):
    from torchvision import get_image_backend
    if get_image_backend() == 'accimage':
        return accimage_loader(path)
    else:
        return pil_loader(path)



# -*- coding: utf-8 -*-
"""
    boris.cli
    ~~~~~~~~~

    A simple command line tool to download samples from
    your datasets on the WhatToLabel platform.

"""

import os
import shutil

import boris.data as data
import hydra
from boris.api import get_samples_by_tag
from boris.cli._helpers import fix_input_path
from tqdm import tqdm


def _download_cli(cfg, is_cli_call=True):
    '''Download all samples in a given tag.
       If from_folder and to_folder are specified, a copy of all images
       in the tag will be stored in the to_folder directory.

    Args:
        cfg['from_folder']: (str, optional)
            Path to folder which holds all images from the dataset
        cfg['to_folder']: (str, optional)
            Path to folder where the copied images will be stored
        cfg['tag_name']: (str) Name of the requested tag
        cfg['dataset_id']: (str) Dataset identifier on the platform
        cfg['token']: (str) Token which grants access to the platform

    '''

    # get all samples in the queried tag
    samples = get_samples_by_tag(
        cfg['tag_name'],
        cfg['dataset_id'],
        cfg['token'],
        mode='list',
        filenames=None
    )

    # store sample names in a .txt file
    with open(cfg['tag_name'] + '.txt', 'w') as f:
        for item in samples:
            f.write("%s\n" % item)

    if cfg['from_folder'] and cfg['to_folder']:
        # "name.jpg" -> "/name.jpg" to prevent bugs like this:
        # "path/to/1234.jpg" ends with both "234.jpg" and "1234.jpg"
        samples = [os.path.join(' ', s)[1:] for s in samples]

        # copy all images from one folder to the other
        from_folder = fix_input_path(cfg['from_folder'])
        to_folder = fix_input_path(cfg['to_folder'])

        dataset = data.BorisDataset(from_folder=from_folder)
        basenames = dataset.get_filenames()

        source_names = [os.path.join(from_folder, f) for f in basenames]
        target_names = [os.path.join(to_folder, f) for f in basenames]

        # only copy files which are in the tag
        indices = [i for i in range(len(source_names))
                   if any([source_names[i].endswith(s) for s in samples])]

        print(f'Copying files from {from_folder} to {to_folder}.')
        for i in tqdm(indices):
            os.makedirs(target_names[i], exist_ok=True)
            shutil.copy(source_names[i], target_names[i])


@hydra.main(config_path='config/config.yaml', strict=False)
def download_cli(cfg):
    '''Download all samples in a given tag.
       If from_folder and to_folder are specified, a copy of all images
       in the tag will be stored in the to_folder directory.

    Args:
        cfg['from_folder']: (str, optional)
            Path to folder which holds all images from the dataset
        cfg['to_folder']: (str, optional)
            Path to folder where the copied images will be stored
        cfg['tag_name']: (str) Name of the requested tag
        cfg['dataset_id']: (str) Dataset identifier on the platform
        cfg['token']: (str) Token which grants access to the platform

    '''
    _download_cli(cfg)


def entry():
    download_cli()



import unittest
import os
import shutil
import torchvision
import tempfile
from boris.data import BorisDataset


class TestBorisDataset(unittest.TestCase):

    def ensure_dir(self, path_to_folder: str):
        if not os.path.exists(path_to_folder):
            os.makedirs(path_to_folder)

    def setUp(self):
        self.available_dataset_names = ['cifar10',
                                        #'cifar100',
                                        #'cityscapes',
                                        #'stl10',
                                        #'voc07-seg',
                                        #'voc12-seg',
                                        #'voc07-det',
                                        #'voc12-det]
                                        ]

    def create_dataset(self, n_subfolders=5, n_samples_per_subfolder=20):
        n_tot = n_subfolders * n_samples_per_subfolder
        dataset = torchvision.datasets.FakeData(size=n_tot,
                                                image_size=(3, 32, 32))

        tmp_dir = tempfile.mkdtemp()

        folder_names = [f'folder_{i}' for i in range(n_subfolders)]
        sample_names = [f'img_{i}.jpg' for i in range(n_samples_per_subfolder)]

        for folder_idx in range(n_subfolders):
            for sample_idx in range(n_samples_per_subfolder):
                idx = (folder_idx * n_subfolders) + sample_idx
                data = dataset[idx]

                self.ensure_dir(os.path.join(tmp_dir,
                                             folder_names[folder_idx]))

                data[0].save(os.path.join(tmp_dir,
                                          folder_names[folder_idx],
                                          sample_names[sample_idx]))
        return tmp_dir, folder_names, sample_names

    def test_create_boris_dataset_from_folder(self):
        n_subfolders = 5
        n_samples_per_subfolder = 10
        n_tot_files = n_subfolders * n_samples_per_subfolder

        dataset_dir, folder_names, sample_names = self.create_dataset(
            n_subfolders,
            n_samples_per_subfolder
        )

        dataset = BorisDataset(from_folder=dataset_dir)
        filenames = dataset.get_filenames()

        fnames = []
        for dir_name in folder_names:
            for fname in sample_names:
                fnames.append(os.path.join(dir_name, fname))

        self.assertEqual(len(filenames), n_tot_files)
        self.assertEqual(len(dataset), n_tot_files)
        self.assertListEqual(sorted(fnames), sorted(filenames))

        shutil.rmtree(dataset_dir)

    def test_create_boris_dataset_from_folder_nosubdir(self):

        # create a dataset
        n_tot = 100
        dataset = torchvision.datasets.FakeData(size=n_tot,
                                                image_size=(3, 32, 32))

        tmp_dir = tempfile.mkdtemp()
        sample_names = [f'img_{i}.jpg' for i in range(n_tot)]
        for sample_idx in range(n_tot):

            data = dataset[sample_idx]
            path = os.path.join(tmp_dir, sample_names[sample_idx])
            data[0].save(path)

        # create boris dataset
        dataset = BorisDataset(from_folder=tmp_dir)
        filenames = dataset.get_filenames()

        # tests
        self.assertEqual(len(filenames), n_tot)
        self.assertEqual(len(dataset), n_tot)
        self.assertListEqual(sorted(sample_names), sorted(filenames))

        for i in range(n_tot):
            sample, target, fname = dataset[i]

    def test_create_boris_dataset_from_torchvision(self):
        tmp_dir = tempfile.mkdtemp()

        for dataset_name in self.available_dataset_names:
            dataset = BorisDataset(root=tmp_dir, name=dataset_name)
            self.assertIsNotNone(dataset)

    def test_not_existing_torchvision_dataset(self):
        list_of_non_existing_names = [
            'a-random-dataset',
            'cifar-100',
            'googleset_ 200'
        ]
        tmp_dir = tempfile.mkdtemp() 
        for dataset_name in list_of_non_existing_names:
            with self.assertRaises(ValueError):
                BorisDataset(root=tmp_dir, name=dataset_name)

    def test_not_existing_folder_dataset(self):
        with self.assertRaises(ValueError):
            BorisDataset(
                from_folder='/a-random-hopefully-non/existing-path-to-nowhere/'
            )



""" ResNet for Self-Supervised Learning """

import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    """ ResNet Basic Block

    """
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        """ Constructor

        Args:
            in_planes: (int) Number of input channels
            planes: (int) Number of channels
            stride: (int) Stride of the first convolutional
        """

        super(BasicBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(planes)

        self.conv2 = nn.Conv2d(planes,
                               planes,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes,
                          self.expansion*planes,
                          kernel_size=1,
                          stride=stride,
                          bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        """ Forward pass through basic ResNet block

        Args:
            x: (Tensor) bsz x channels x W x H

        Returns:
            out: (Tensor) bsz x channels x W x H
        """

        out = self.conv1(x)
        out = self.bn1(out)
        out = F.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += self.shortcut(x)
        out = F.relu(out)

        return out


class Bottleneck(nn.Module):
    """ ResNet Bottleneck block

    """
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        """ Constructor

        Args:
            in_planes: (int) Number of input channels
            planes: (int) Number of intermediate channels
            stride: (int) Stride of first convolutional
        """

        super(Bottleneck, self).__init__()

        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)

        self.conv2 = nn.Conv2d(planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.conv3 = nn.Conv2d(planes,
                               self.expansion*planes,
                               kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes,
                          self.expansion*planes,
                          kernel_size=1,
                          stride=stride,
                          bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        """ Forward pass through bottleneck ResNet block

        Args:
            x: (Tensor) bsz x channels x W x H

        Returns:
            out: (Tensor) bsz x (expansion*channels) x W x H
        """

        out = self.conv1(x)
        out = self.bn1(out)
        out = F.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = F.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += self.shortcut(x)
        out = F.relu(out)

        return out


class ResNet(nn.Module):
    """ ResNet

    Reference:
        [1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
        Deep Residual Learning for Image Recognition. arXiv:1512.03385
    """

    def __init__(self,
                 block=BasicBlock,
                 layers=[2, 2, 2, 2],
                 num_classes=10,
                 width=1):
        """ Constructor

        Args:
            block: (nn.Module) ResNet building block type
            layers: (List[int]) Blocks per layer
            num_classes: (int) Number of class for final softmax
            width: (int) Multiplier for ResNet width

        """

        super(ResNet, self).__init__()
        self.in_planes = int(64 * width)

        self.base = int(64 * width)

        self.conv1 = nn.Conv2d(3,
                               self.base,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(self.base)
        self.layer1 = self._make_layer(block, self.base, layers[0], stride=1)
        self.layer2 = self._make_layer(block, self.base*2, layers[1], stride=2)
        self.layer3 = self._make_layer(block, self.base*4, layers[2], stride=2)
        self.layer4 = self._make_layer(block, self.base*8, layers[3], stride=2)
        self.linear = nn.Linear(self.base*8*block.expansion, num_classes)

    def _make_layer(self, block, planes, layers, stride):
        strides = [stride] + [1]*(layers-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def ResNetGenerator(name='resnet-18', width=1, num_classes=10):
    """ Builds and returns the specified ResNet

    Args:
        name: (str) ResNet version from resnet-{9, 18, 34, 50, 101, 152}

    Returns:
        ResNet as nn.Module
    """

    model_params = {
        'resnet-9': {'block': BasicBlock, 'layers': [1, 1, 1, 1]},
        'resnet-18': {'block': BasicBlock, 'layers': [2, 2, 2, 2]},
        'resnet-34': {'block': BasicBlock, 'layers': [3, 4, 6, 3]},
        'resnet-50': {'block': Bottleneck, 'layers': [3, 4, 6, 3]},
        'resnet-101': {'block': Bottleneck, 'layers': [3, 4, 23, 3]},
        'resnet-152': {'block': Bottleneck, 'layers': [3, 8, 36, 3]},
    }

    if name not in model_params.keys():
        raise ValueError('Illegal name: {%s}. \
        Try resnet-9, resnet-18, resnet-34, resnet-50, resnet-101, resnet-152.' % (name))

    return ResNet(**model_params[name], width=width, num_classes=10)


class ResNetSimCLR(nn.Module):
    """ Self-supervised model with ResNet architecture

    """

    def __init__(self, name='resnet-18', width=1, num_ftrs=16, out_dim=128):
        """ Constructor

        Args:
            name: (str) ResNet version from resnet-{9, 18, 34, 50, 101, 152}
            num_ftrs: (int) Embedding dimension
            out_dim: (int) Output dimension

        """

        self.num_ftrs = num_ftrs
        self.out_dim = out_dim

        super(ResNetSimCLR, self).__init__()
        resnet = ResNetGenerator(name=name, width=width)

        last_conv_channels = list(resnet.children())[-1].in_features

        self.features = nn.Sequential(
            nn.BatchNorm2d(3),
            *list(resnet.children())[:-1],
            nn.Conv2d(last_conv_channels, num_ftrs, 1),
            nn.AdaptiveAvgPool2d(1)
        )

        self.projection_head = nn.Sequential(
            nn.Linear(num_ftrs, num_ftrs),
            nn.ReLU(),
            nn.Linear(num_ftrs, out_dim)
        )

    def forward(self, x):
        """ Forward pass through ResNetSimCLR

        Args:
            x: (tensor) bsz x channels x W x H

        Returns:
            embeddings: (tensor) bsz x n_features
        """
        # embed images in feature space
        emb = self.features(x)
        emb = emb.squeeze()

        # return projection to TODO
        return self.projection_head(emb)



import torch
import numpy as np

"""
def _evaluate(dataloader, model, metric, device=None):

    with torch.no_grad():
        for batch, target, _ in dataloader:

            batch = batch.to(device)
            target = target.to(device)

            output = model(batch)
            # metric automatically updates all necessary
            # numbers and statistics
            metric(output, target, len(dataloader))

    return metric
"""


def _evaluate(dataloader, model, metrics, device=None):

    with torch.no_grad():
        for batch, target, _ in dataloader:

            batch = batch.to(device)
            target = target.to(device)

            output = model(batch)

            for i, metric in enumerate(metrics):
                metric(output, target, len(dataloader))
                assert np.power(metric.result - metrics[i].result, 2) < 1e-5

    return metrics


class Evaluatable(object):

    def __init__(self, dataloader, eval_metrics):
        self.dataloader = dataloader
        self.eval_metrics = eval_metrics

    def evaluate(self, trainable, device=None):
        """Evaluate trainable with different metrics.

        """
        self.eval_metrics = _evaluate(
            self.dataloader, trainable, self.eval_metrics, device=device
        )



""" boris.cli

    The boris.cli module provides a console interface
    for training self-supervised models, embedding,
    and filtering datasets

"""

from boris import is_lightning_available

if is_lightning_available():
    from .train_cli import train_cli
    from .embed_cli import embed_cli
    from .boris_cli import boris_cli

from .upload_cli import upload_cli



import setuptools
import sys


if __name__ == '__main__':

    with open('README.md') as f:
        long_description = f.read()

    LOCAL_INSTALL = False
    if '--local' in sys.argv:
        sys.argv.remove('--local')
        LOCAL_INSTALL = True

    name = 'borisml'
    version = '0.1.4'
    author = 'Philipp Wirth & Igor Susmelj'
    author_email = 'philipp@whattolabel.com'
    description = "A deep learning package for self-supervised learning"

    entry_points = {
        "console_scripts": [
            "boris-train = boris.cli.train_cli:entry [LIGHTNING]",
            "boris-embed = boris.cli.embed_cli:entry [LIGHTNING]",
            "boris-magic = boris.cli.boris_cli:entry [LIGHTNING]",
            "boris-upload = boris.cli.upload_cli:entry",
            "boris-download = boris.cli.download_cli:entry",
        ]
    }

    install_requires = [
        'tqdm',
        'torchvision',
        'pandas>=1.0.3',
        'hydra-core>=0.11.3',
        'numpy>=1.18.1',
        'requests>=2.23.0',
    ]

    extras = {
        'LIGHTNING': [
            'pytorch_lightning>=0.7.1',
            'opencv-python'
        ]
    }

    packages = [
        'boris',
        'boris.api',
        'boris.cli',
        'boris.data',
        'boris.embedding',
        'boris.loss',
        'boris.models',
        'boris.transforms',
        'boris.utils'
    ]

    project_urls = {
        'Documentation': 'https://www.notion.so/whattolabel/WhatToLabel-Documentation-28e645f5564a453e807d0a384a4e6ea7',
        'WhatToLabel': 'https://www.whattolabel.com'
    }

    if LOCAL_INSTALL:
        packages.append(
            'boris.sampling'
        )

    classifiers = [
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License"
    ]

    setuptools.setup(
        name=name,
        version=version,
        author=author,
        author_email=author_email,
        description=description,
        entry_points=entry_points,
        extras_require=extras,
        license='MIT',
        long_description=long_description,
        long_description_content_type='text/markdown',
        install_requires=install_requires,
        packages=packages,
        classifiers=classifiers,
        include_package_data=True,
        project_urls=project_urls,
    )



""" Helper Functions """

import os

import torchvision.datasets as datasets
from boris.data._image_loaders import default_loader

IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp',
                  '.pgm', '.tif', '.tiff', '.webp')


def _make_dataset(directory, extensions=None, is_valid_file=None):
    """Return a list of all image files with targets in the directory

    Args:
        directory: (str) Root directory path
            (should not contain subdirectories!)
        extensions: (List[str]) List of allowed extensions
        is_valid_file: (callable) Used to check corrupt files

    Returns:
        List of instance tuples: (path_i, target_i = 0)

    """

    if extensions is not None:
        def _is_valid_file(filename):
            return filename.lower().endswith(extensions)

    instances = []
    for fname in os.listdir(directory):

        if not _is_valid_file(fname):
            continue

        path = os.path.join(directory, fname)
        item = (path, 0)
        instances.append(item)

    return instances


class DatasetFolder(datasets.VisionDataset):

    def __init__(self, root, loader, extensions=None, transform=None,
                 target_transform=None, is_valid_file=None):
        """Constructor based on torchvisions DatasetFolder
            (https://pytorch.org/docs/stable/torchvision/datasets.html#datasetfolder)

        Args:
            root: (str) Root directory path
            loader: (callable) Function that loads file at path
            extensions: (List[str]) List of allowed extensions
            transform: Function that takes a PIL image and returns
                transformed version
            target_transform: As transform but for targets
            is_valid_file: (callable) Used to check corrupt files

        Raises:
            RuntimeError: If no supported files are found in root.

        """

        super(DatasetFolder, self).__init__(root, transform=transform,
                                            target_transform=target_transform)

        samples = _make_dataset(self.root, extensions, is_valid_file)
        if len(samples) == 0:
            msg = 'Found 0 files in folder: {}\n'.format(self.root)
            if extensions is not None:
                msg += 'Supported extensions are: {}'.format(
                    ','.join(extensions))
            raise RuntimeError(msg)

        self.loader = loader
        self.extensions = extensions

        self.samples = samples
        self.targets = [s[1] for s in samples]

    def __getitem__(self, index):
        """Returns item at index

        Args:
            index: (int) Index

        Returns:
            tuple: (sample, target) where target is 0

        """

        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target

    def __len__(self):
        return len(self.samples)


def _contains_subdirs(root):
    """Check whether directory contains subdirectories

    Args:
        root: (str) Root directory path

    Returns:
        True if root contains subdirectories else false

    """

    list_dir = os.listdir(root)
    is_dir = [
        os.path.isdir(os.path.join(root, f)) for f in list_dir
    ]
    return any(is_dir)


def _load_dataset_from_folder(root, transform):
    """Initialize dataset from folder

    Args:
        root: (str) Root directory path
        transform: (torchvision.transforms.Compose) image transformations

    Returns:
        Dataset consisting of images in the root directory
    """

    if _contains_subdirs(root):
        dataset = datasets.ImageFolder(root, transform=transform)
    else:
        dataset = DatasetFolder(root, default_loader,
                                extensions=IMG_EXTENSIONS,
                                transform=transform)

    return dataset


def _load_dataset(root='',
                  name='cifar10',
                  train=True,
                  download=True,
                  transform=None,
                  from_folder=''):
    """ Initialize dataset from torchvision or from folder

    Args:
        root: (str) Directory where dataset is stored
        name: (str) Name of the dataset (e.g. cifar10, cifar100)
        train: (bool) Use the training set
        download: (bool) Download the dataset
        transform: (torchvision.transforms.Compose) image transformations
        from_folder: (str) Path to directory holding the images to load.

    Returns:
        A torchvision dataset

    Raises:
        ValueError: If the specified dataset doesn't exist

    """

    if from_folder and os.path.exists(from_folder):
        # load data from directory
        dataset = _load_dataset_from_folder(from_folder,
                                            transform)

    elif name.lower() == 'cifar10' and root:
        # load cifar10
        dataset = datasets.CIFAR10(root,
                                   train=train,
                                   download=download,
                                   transform=transform)

    elif name.lower() == 'cifar100' and root:
        # load cifar100
        dataset = datasets.CIFAR100(root,
                                    train=train,
                                    download=download,
                                    transform=transform)

    elif name.lower() == 'cityscapes' and root:
        # load cityscapes
        root = os.path.join(root, 'cityscapes/')
        split = 'train' if train else 'val'
        dataset = datasets.Cityscapes(root,
                                      split=split,
                                      transform=transform)

    elif name.lower() == 'stl10' and root:
        # load stl10
        split = 'train' if train else 'test'
        dataset = datasets.STL10(root,
                                 split=split,
                                 download=download,
                                 transform=transform)

    elif name.lower() == 'voc07-seg' and root:
        # load pascal voc 07 segmentation dataset
        image_set = 'train' if train else 'val'
        dataset = datasets.VOCSegmentation(root,
                                           year='2007',
                                           image_set=image_set,
                                           download=download,
                                           transform=transform)

    elif name.lower() == 'voc12-seg' and root:
        # load pascal voc 12 segmentation dataset
        image_set = 'train' if train else 'val'
        dataset = datasets.VOCSegmentation(root,
                                           year='2012',
                                           image_set=image_set,
                                           download=download,
                                           transform=transform)

    elif name.lower() == 'voc07-det' and root:
        # load pascal voc 07 object detection dataset
        image_set = 'train' if train else 'val'
        dataset = datasets.VOCDetection(root,
                                        year='2007',
                                        image_set=image_set,
                                        download=True,
                                        transform=transform)

    elif name.lower() == 'voc12-det' and root:
        # load pascal voc 12 object detection dataset
        image_set = 'train' if train else 'val'
        dataset = datasets.VOCDetection(root,
                                        year='2012',
                                        image_set=image_set,
                                        download=True,
                                        transform=transform)

    else:
        raise ValueError(
            'The specified dataset or datafolder does not exist (%s, %s)'
            % (name, from_folder))

    return dataset



from typing import List

import numpy as np
import pandas as pd


def embeddings_to_pandas(embeddings: np.ndarray,
                         labels: List[int],
                         filenames: List[str]) -> pd.DataFrame:
    """Creates pandas dataframe from embeddings, labels and filenames

    Args:
        embeddings: 2-dimensional numpy array containing embeddings
        labels: List of label number (e.g. different subfolders)
        filenames: List of filenames

    Returns:
        Pandas dataframe ready to be saved to csv
    """
    all_equal_length = True
    all_equal_length = all_equal_length and (len(embeddings) == len(labels))
    all_equal_length = all_equal_length and (len(labels) == len(filenames))

    if not all_equal_length:
        raise ValueError(f'Length of embeddings {len(embeddings)} \
            labels ({len(labels)}) and filenames ({len(filenames)}) \
            do not match!')
    df = pd.DataFrame.from_records(embeddings)
    df.index = filenames
    df["label"] = labels
    df = df.reset_index()
    cols = ['filename']
    cols += [f'embedding_{i}' for i in range(embeddings.shape[1])]
    cols += ['label']
    df.columns = cols
    return df


def pandas_to_dict(df: pd.DataFrame) -> dict:
    """Converts pandas dataframe into dictionary with proper format for uploading

    Args:
        df: Pandas dataframe with embeddings and filenames

    Returns:
        A dictionary
    """
    columns = list(df)
    embedding_columns = list(filter(lambda x: 'embedding' in x, columns))
    embeddings = df[embedding_columns].values.tolist()
    filenames = [row for row in df['filename']]
    labels = [row for row in df['label']]
    data = {}
    data['embeddingName'] = 'default'
    data['embeddings'] = []
    for (embedding, fname, label) in zip(embeddings, filenames, labels):
        item = {}
        item['fileName'] = fname
        item['value'] = embedding
        item['label'] = label
        data['embeddings'].append(item)

    return data



""" Collate Functions """

import torch
import torch.nn as nn

import torchvision.transforms as transforms
from boris.transforms import GaussianBlur


class BaseCollateFunction(nn.Module):
    """ Base class for collate callables.

    """

    def __init__(self, transform):
        """ Constructor

        Args:
            transform: (torchvision.transforms.Compose)

        """
        super(BaseCollateFunction, self).__init__()
        self.transform = transform

    def forward(self, batch):
        """ Upon call

            Args:
                batch: List[(img, target, fname)]

            Returns:
                image_batch: Transformed images in batch form
                target_batch: Targets in batch form
                fnames: List of filenames for samples in batch
        """
        batch_a = torch.cat(
            [self.transform(item[0]).unsqueeze_(0) for item in batch], 0
        )
        batch_b = torch.cat(
            [self.transform(item[0]).unsqueeze_(0) for item in batch], 0
        )
        labels = torch.tensor([item[1] for item in batch]).long()
        fnames = [item[2] for item in batch]
        return (torch.cat((batch_a, batch_b), 0),
                torch.cat((labels, labels), 0),
                fnames)


class ImageCollateFunction(BaseCollateFunction):
    """ Image collate function for self-supervised learning

    """

    def __init__(self, input_size=32, cj_prob=0.8, cj_bright=0.7,
                 cj_contrast=0.7, cj_sat=0.7, cj_hue=0.2, min_scale=0.08,
                 random_gray_scale=0.2, kernel_size=0.1):
        """ Constructor

        Args:
            input_size: (int) Dimensions of the input image
            cj_prob: (float) Probability that color jitter is applied
            cj_bright: (float) How much to jigger brightness
            cj_contrast: (float) How much to jigger constrast
            cj_sat: (float) How much to jigger saturation
            cj_hue: (float) How much to jigger hue
            min_scale: (float) Minimum size of the randomized crop
            random_gray_scale: (float) Probability of grayscale
            kernel_size: (float) Sigma of gaussian blur is \
                kernel_size * input_size

        """
        color_jitter = transforms.ColorJitter(
            cj_bright, cj_contrast, cj_sat, cj_hue
        )
        transform = transforms.Compose(
            [transforms.RandomResizedCrop(size=input_size,
                                          scale=(min_scale, 1.0)),
             transforms.RandomHorizontalFlip(),
             transforms.RandomApply([color_jitter], p=cj_prob),
             transforms.RandomGrayscale(p=random_gray_scale),
             GaussianBlur(kernel_size=int(kernel_size * input_size)),
             transforms.ToTensor(),
             transforms.Normalize(
                mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]),
             ]
        )
        super(ImageCollateFunction, self).__init__(transform)



import os


import pandas as pd

from itertools import islice
from typing import List

from boris.api._communication import create_initial_tag
from boris.api._communication import get_presigned_upload_url
from boris.api._communication import upload_file_with_signed_url
from boris.api._communication import upload_embedding
from boris.api._communication import get_samples
from boris.data import BorisDataset
from boris.utils import pandas_to_dict

import tqdm
import warnings
from concurrent.futures import ThreadPoolExecutor

MAXIMUM_DATASET_SIZE = 10_000


def upload_embeddings_from_csv(path_to_embeddings: str,
                               dataset_id: str,
                               token: str,
                               max_upload: int = 512):
    """Uploads embeddings from a csv file to the cloud solution

    Args:
        path_to_embeddings: Path to csv file containing embeddings
        dataset_id: The unique identifier for the dataset
        token: Token for authentication

    Raises:
        RuntimeError if there is an error during the upload
        of a batch of embeddings
    """
    df = pd.read_csv(path_to_embeddings)
    data = pandas_to_dict(df)

    for embedding in data['embeddings']:
        embedding['fileName'] = os.path.basename(embedding['fileName'])

    data['token'] = token
    data['datasetId'] = dataset_id

    def _upload_single_batch(batch, append=False):
        batch['append'] = int(append)
        upload_embedding(batch)

    n_embeddings = len(data['embeddings'])
    n_batches = n_embeddings // max_upload
    n_batches = n_batches + 1 if n_embeddings % max_upload else n_batches

    embedding_batches = [None] * n_batches
    for i in range(n_batches):
        left = i*max_upload
        right = min((i + 1) * max_upload, n_embeddings)
        batch = data.copy()
        batch['embeddings'] = data['embeddings'][left:right]
        embedding_batches[i] = batch

    pbar = tqdm.tqdm(unit='embs', total=n_embeddings)
    for i, batch in enumerate(embedding_batches):
        _upload_single_batch(batch, append=i > 0)
        pbar.update(len(batch['embeddings']))


def upload_images_from_folder(path_to_folder: str,
                              dataset_id: str,
                              token: str,
                              max_workers: int = 8,
                              max_requests: int = 8):
    """Uploads images from a directory to the cloud solution

    Args:
        path_to_folder: Path to the folder containing the images
        dataset_id: The unique identifier for the dataset
        token: Token for authentication

    """

    bds = BorisDataset(from_folder=path_to_folder)
    fnames = bds.get_filenames()

    if len(fnames) > MAXIMUM_DATASET_SIZE:
        msg = f'Your dataset has {len(fnames)} samples which'
        msg += f'is more than the allowed maximum of {MAXIMUM_DATASET_SIZE}'
        raise ValueError(msg)

    def _upload_single_image(fname):

        success = True
        # try to get signed url
        try:
            basename = fname
            signed_url = get_presigned_upload_url(
                basename, dataset_id, token
            )
        except RuntimeError:
            success = False

        # try to upload image
        try:
            if success:
                full_fname = os.path.join(path_to_folder, fname)
                upload_file_with_signed_url(full_fname, signed_url)
        except RuntimeError:
            success = False

        return success

    n_batches = len(fnames) // max_requests
    n_batches = n_batches + 1 if len(fnames) % max_requests else n_batches
    fname_batches = [
        list(islice(fnames, i * max_requests, (i + 1) * max_requests))
        for i in range(n_batches)
    ]

    chunksize = max(max_requests // max_workers, 1)
    executor = ThreadPoolExecutor(max_workers=max_workers)

    pbar = tqdm.tqdm(unit='imgs', total=len(fnames))
    for i, batch in enumerate(fname_batches):
        mapped = executor.map(_upload_single_image, batch, chunksize=chunksize)
        mapped = list(mapped)
        if not all(mapped):
            msg = 'Warning: Unsuccessful upload(s) in batch {}! '.format(i)
            msg += 'This could cause problems when uploading embeddings.'
            msg += 'Failed at file: {}'.format(mapped.index(False))
            warnings.warn(msg)
        pbar.update(len(batch))
    create_initial_tag(dataset_id, token)


def upload_metadata_from_json(path_to_embeddings: str,
                              dataset_id: str,
                              token: str):
    pass


def get_samples_by_tag(tag_name: str,
                       dataset_id: str,
                       token: str,
                       mode: str = 'list',
                       filenames: List[str] = None):
    """Get the files associated with a given tag and dataset

    Args:
        tag_name: Name of the tag to query
        dataset_id: The unique identifier of the dataset
        token: Token for authentication
        mode: Return type, must be in ["list", "mask", "indices"]
        filenames: List of all filenames

    Returns:
        Either list of filenames, binary mask, or list of indices
        specifying the samples in the requested tag

    Raises: ValueError, RuntimeError

    """

    if mode == 'mask' and filenames is None:
        msg = f'Argument filenames must not be None for mode "{mode}"!'
        raise ValueError(msg)
    if mode == 'indices' and filenames is None:
        msg = f'Argument filenames must not be None for mode "{mode}"!'
        raise ValueError(msg)

    samples = get_samples(dataset_id, token, tag_name=tag_name)

    if mode == 'list':
        return samples

    if mode == 'mask':
        mask = [1 if f in set(samples) else 0 for f in filenames]
        if sum(mask) != len(samples):
            msg = 'Error during mapping from samples to filenames: '
            msg += f'sum(mask) != len(samples) with lengths '
            msg += f'{sum(mask)} and {len(samples)}'
            raise RuntimeError(msg)
        return mask

    if mode == 'indices':
        indices = [i for i in range(len(filenames))]
        indices = filter(lambda i: filenames[i] in set(samples), indices)
        indices = list(indices)
        if len(indices) != len(samples):
            msg = 'Error during mapping from samples to filenames: '
            msg += f'len(indices) != len(samples) with lengths '
            msg += f'{len(indices)} and {len(samples)}.'
            raise RuntimeError(msg)
        return indices

    msg = f'Got illegal mode "{mode}"! '
    msg += 'Must be in ["list", "mask", "indices"]'
    raise ValueError(msg)



from boris.cli.train_cli import _train_cli
from boris.cli.embed_cli import _embed_cli
from boris.cli.boris_cli import _boris_cli
import boris.cli as cli

import yaml
import os


def _get_config_path(config_path):
    """Find path to yaml config file

    Args:
        config_path: (str) Path to config.yaml file

    Returns:
        Path to config.yaml if specified else default config.yaml

    Raises:
        ValueError if the config_path is not None but doesn't exist

    """
    if config_path is None:
        dirname = os.path.dirname(cli.__file__)
        config_path = os.path.join(dirname, 'config/config.yaml')
    if not os.path.exists(config_path):
        raise ValueError("Config path {} does not exist!".format(config_path))

    return config_path


def _load_config_file(config_path):
    """Load a yaml config file

    Args:
        config_path: (str) Path to config.yaml file

    Returns:
        Dictionary with configs from config.yaml

    """
    Loader = yaml.FullLoader
    with open(config_path, 'r') as config_file:
        cfg = yaml.load(config_file, Loader=Loader)

    return cfg


def _add_kwargs(cfg, kwargs):
    """Add keyword arguments to config

    Args:
        cfg: (dict) Dictionary of configs from config.yaml
        kwargs: (dict) Dictionary of keyword arguments

    Returns:
        Union of cfg and kwargs

    """
    for key, item in kwargs.items():
        if isinstance(item, dict):
            if key in cfg:
                cfg[key] = _add_kwargs(cfg[key], item)
            else:
                cfg[key] = item
        else:
            cfg[key] = item

    return cfg


def train_model_and_get_image_features(config_path: str = None, **kwargs):
    """Train a self-supervised model and use it to embed images

    Args:
        config_path: Path to config.yaml
        **kwargs: Keyword arguments to overwrite configs

    Returns:
        (embeddings, labels, filenames)

    """
    config_path = _get_config_path(config_path)
    config_args = _load_config_file(config_path)
    config_args = _add_kwargs(config_args, kwargs)

    return _boris_cli(config_args, is_cli_call=False)


def train_self_supervised_model(config_path: str = None, **kwargs):
    """Train a self-supervised model

    Args:
        config_path: Path to config.yaml
        **kwargs: Keyword arguments to overwrite configs

    Returns:
        Path to embedding-checkpoint

    """
    config_path = _get_config_path(config_path)
    config_args = _load_config_file(config_path)
    config_args = _add_kwargs(config_args, kwargs)

    return _train_cli(config_args, is_cli_call=False)


def get_image_features(checkpoint: str, config_path: str = None, **kwargs):
    """Embed images with a self-supervised model.

    Args:
        config_path: Path to config.yaml
        **kwargs: Keyword arguments to overwrite configs

    Returns:
        (embeddings, labels, filenames):

    """
    config_path = _get_config_path(config_path)
    config_args = _load_config_file(config_path)
    config_args = _add_kwargs(config_args, kwargs)

    config_args['checkpoint'] = checkpoint

    return _embed_cli(config_args, is_cli_call=False)



""" Gaussian Blur """
# Copyright (c) 2020. Mirage Technologies AG and its affiliates.
# All Rights Reserved

import warnings

import numpy as np
from boris import is_opencv_available

if is_opencv_available():
    import cv2
else:
    msg = 'cv2 is not available. Will skip Gaussian Blur augmentations.'
    warnings.warn(msg)


class GaussianBlur(object):
    # Implements Gaussian blur as described in the SimCLR paper
    def __init__(self, kernel_size, min=0.1, max=2.0, prob=0.5):
        self.min = min
        self.max = max
        self.prob = prob
        # kernel size is set to be 10% of the image height/width
        self.kernel_size = kernel_size
        if self.kernel_size % 2 == 0:
            self.kernel_size += 1

    def __call__(self, sample):
        sample = np.array(sample)

        if not is_opencv_available():
            return sample

        # blur the image with a 50% chance
        prob = np.random.random_sample()
        if prob < self.prob:
            sigma = (self.max - self.min) * \
                np.random.random_sample() + self.min

            sample = cv2.GaussianBlur(sample,
                                      (self.kernel_size, self.kernel_size),
                                      sigma)

        return sample



""" BorisDataset """

import os
from typing import List

import torch.utils.data as data
import torchvision.datasets as datasets

from boris.data._helpers import _load_dataset
from boris.data._helpers import DatasetFolder


class BorisDataset(data.Dataset):
    """ Provides a uniform data interface for the embedding models.

    """

    def __init__(self, root='',
                 name='cifar10',
                 train=True,
                 download=True,
                 transform=None,
                 from_folder=''):
        """ Constructor

        Args:
            root: (str) Directory where dataset is stored
            name: (str) Name of the dataset (e.g. cifar10, cifar100)
            train: (bool) Use the training set
            download: (bool) Download the dataset
            transform: (torchvision.transforms.Compose) image transformations
            from_folder: (str) Path to directory holding the images to load.

        Raises:
            ValueError: If the specified dataset doesn't exist

        """

        super(BorisDataset, self).__init__()
        self.dataset = _load_dataset(
            root, name, train, download, transform, from_folder
        )
        self.root_folder = None
        if from_folder:
            self.root_folder = from_folder

    def get_filenames(self) -> List[str]:
        """Returns a list of filenames
        """
        list_of_filenames = []
        for index in range(len(self)):
            fname = self._get_filename_by_index(index)
            list_of_filenames.append(fname)
        return list_of_filenames

    def _get_filename_by_index(self, index) -> str:
        """Returns filename based on index
        """
        if isinstance(self.dataset, datasets.ImageFolder):
            full_path = self.dataset.imgs[index][0]
            return os.path.relpath(full_path, self.root_folder)
        elif isinstance(self.dataset, DatasetFolder):
            full_path = self.dataset.samples[index][0]
            return os.path.relpath(full_path, self.root_folder)
        else:
            return str(index)

    def __getitem__(self, index):
        """ Get item at index. Supports torchvision.ImageFolder datasets and
            all dataset which return the tuple (sample, target).

        Args:
         - index:   index of the queried item

        Returns:
         - sample:  sample at queried index
         - target:  class_index of target class, 0 if there is no target
         - fname:   filename of the sample, str(index) if there is no filename

        """
        fname = self._get_filename_by_index(index)
        sample, target = self.dataset.__getitem__(index)
        return sample, target, fname

    def __len__(self):
        """ Length of the dataset
        """
        return len(self.dataset)

    def __add__(self, other):
        """ Add element to dataset
        """
        raise NotImplementedError()



""" Contrastive Loss Functions """

import torch
import numpy as np


class NTXentLoss(torch.nn.Module):
    """ Contrastive Cross Entropy Loss """

    def __init__(self, temperature=0.5, use_cosine_similarity=True):
        """ Constructor

        Args:
            temperature: (float) Scale logits
            use_cosine_similarity: (bool) Use cosine similarity over L2

        Raises:
            ValueError: If abs(temperature) < 1e-8 to prevent divide by zero
        """
        super(NTXentLoss, self).__init__()
        self.temperature = temperature
        self.similarity_function = self._get_similarity_function(
                                        use_cosine_similarity)
        self.cross_entropy = torch.nn.CrossEntropyLoss(reduction="sum")
        self.correlated_mask = None
        self.eps = 1e-8

        if abs(self.temperature) < self.eps:
            raise ValueError('Illegal temperature: abs({}) < 1e-8'
                             .format(self.temperature))

    def _get_similarity_function(self, use_cosine_similarity):
        if use_cosine_similarity:
            self._cosine_similarity = torch.nn.CosineSimilarity(dim=-1)
            return self._cosine_simililarity
        else:
            return self._dot_simililarity

    def _torch_get_correlated_mask(self, batch_size, device=None):
        diag = torch.eye(2 * batch_size, device=device)
        diag[batch_size:, :batch_size] += torch.eye(batch_size, device=device)
        diag[:batch_size, batch_size:] += torch.eye(batch_size, device=device)
        mask = (1 - diag).type(torch.bool)
        return mask

    def _get_correlated_mask(self, batch_size):
        # TODO: deprecate
        diag = np.eye(2 * batch_size)
        l1 = np.eye((2 * batch_size), 2 * batch_size, k=-batch_size)
        l2 = np.eye((2 * batch_size), 2 * batch_size, k=batch_size)
        mask = torch.from_numpy((diag + l1 + l2))
        mask = (1 - mask).type(torch.bool)
        if torch.cuda.is_available():
            mask.to("cuda")
        return mask

    @staticmethod
    def _dot_simililarity(x, y):
        v = torch.tensordot(x.unsqueeze(1), y.T.unsqueeze(0), dims=2)
        # x shape: (N, 1, C)
        # y shape: (1, C, 2N)
        # v shape: (N, 2N)
        return v

    def _cosine_simililarity(self, x, y):
        # x shape: (N, 1, C)
        # y shape: (1, 2N, C)
        # v shape: (N, 2N)
        v = self._cosine_similarity(x.unsqueeze(1), y.unsqueeze(0))
        return v

    def forward(self, output, labels=None):
        """ Upon call

            Args:
                output: (torch.Tensor) Output from the model, shape: 2*bsz x d

            Returns:
                loss: (torch.Tensor) Contrastive Cross Entropy Loss

            Raises:
                ValueError: If shape of output is not multiple of batch_size
        """

        output = torch.nn.functional.normalize(output, dim=1)
        device = output.device

        if output.shape[0] % 2:
            raise ValueError('Illegal shape, axis 0 not divisible by 2: {}'
                             .format(output.shape))

        batch_size = output.shape[0] // 2
        similarity_matrix = self.similarity_function(output, output)

        # filter out the scores from the positive samples
        l_pos = torch.diag(similarity_matrix, batch_size)
        r_pos = torch.diag(similarity_matrix, -batch_size)
        positives = torch.cat([l_pos, r_pos]).view(2 * batch_size, 1)

        if self.correlated_mask is None:
            self.correlated_mask = \
                self._torch_get_correlated_mask(batch_size, device=device)
        if 2 * batch_size != self.correlated_mask.shape[0]:
            self.correlated_mask = \
                self._torch_get_correlated_mask(batch_size, device=device)

        negatives = similarity_matrix[
            self.correlated_mask
            ].view(2 * batch_size, -1)

        logits = torch.cat((positives, negatives), dim=1)
        logits /= (self.temperature + self.eps)

        labels = torch.zeros(2 * batch_size).long()

        loss = self.cross_entropy(logits, labels.to(device))
        return loss / (2 * batch_size)



import os


def _prepend_hydra_prefix(path):
    """Hydra creates a working directory depending on the time stamps.
       The format is:
            outputs/*-*-*/*-*-*/
       This makes it inconvenient to pass relative paths arguments as
       the relative paths will be evaluated from the working directory.

       This function prepends a prefix to the relative path to move out
       of the current working directory into the directory where the cli
       was initially executed. It then turns the relative path into an
       absolute path and returns it.
    """
    path = os.path.join('../../../', path)
    return os.path.abspath(path)


def fix_input_path(path):
    """Fix broken relative paths.

    """
    if not os.path.isabs(path):
        path = _prepend_hydra_prefix(path)
    return path


def filter_state_dict(state_dict):
    """Prevent unexpected key error when loading PyTorch-Lightning checkpoints
       by removing the unnecessary prefix model. from each key.

    """
    new_state_dict = {}
    for key, item in state_dict.items():
        new_key = '.'.join(key.split('.')[1:])
        new_state_dict[new_key] = item
    return new_state_dict


def _is_url(checkpoint):
    is_url = ('https://storage.googleapis.com' in checkpoint)
    return is_url


def _model_zoo(model):

    zoo = {

        'resnet-9/simclr/d16/w0.125':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet9-simclr-d16-w0.125-i-7269c38d.pth',

        'resnet-18/simclr/d16/w1.0':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet18-simclr-d16-w1.0-i-58852cb9.pth',

        'resnet-34/simclr/d16/w1.0':
        'https://storage.googleapis.com/models_boris/whattolabel-resnet34-simclr-d16-w1.0-i-6e80d963.pth',
        
    }

    key = model['name']
    key += '/simclr'
    key += '/d' + str(model['num_ftrs'])
    key += '/w' + str(float(model['width']))

    if key in zoo.keys():
        return zoo[key], key
    else:
        return '', key


